#include "../include/ImageLib.h"
#include "image_util.h"
#include <assert.h>
#include <string.h>


Image::Image( void ) : m_pixels(0)
{
}
Image::Image( const Image &image ) : m_pixels(0)
{
	Copy( &image );
}
Image::Image( unsigned int width, unsigned int height, unsigned char chan_number ) : m_pixels(0)
{
	Create( width, height, chan_number );
}
Image::Image( const char *filename, Image::Format format ) : m_pixels(0)
{
	Load( filename, format );
}


Image::~Image( void )
{
	Clear();
}


// Creation
bool Image::Create( unsigned int width, unsigned int height, unsigned char chan_number )
{
	Clear();

	m_width       = width;
	m_height      = height;
	m_chan_number = chan_number;
	m_depth       = 1;

	Update();

	if( m_pixel_number > 0 )
	{
		m_pixels = new unsigned char[ m_pixel_number * m_chan_mult_depth ];
		memset( m_pixels, 0, m_pixel_number * m_chan_mult_depth * sizeof(unsigned char) );
	}

	return true;
}


// Recopie une image
bool Image::Copy ( const Image *image )
{
	Clear();

	m_width       = image->m_width;
	m_height      = image->m_height;
	m_chan_number = image->m_chan_number;
	m_depth       = image->m_depth;

	Update();

	if( image->IsValid() )
	{
		m_pixels = new unsigned char[ m_pixel_number * m_chan_mult_depth ];
		memcpy( m_pixels, image->m_pixels, m_pixel_number * m_chan_mult_depth * sizeof(unsigned char) );
	}
	else
	{
		m_pixels = 0;
	}

	return true;
}


// Charge une image depuis un fichier
bool Image::Load( const char *filename, Image::Format format )
{
	Clear();

	switch( format )
	{
	case Image::PBM:
	case Image::PGM:
	case Image::PPM:
	case Image::PNM: return LoadPNM( filename );
	case Image::BMP: return LoadBMP( filename );
	case Image::JPG: return LoadJPG( filename );
	case Image::PNG: return LoadPNG( filename );

	case Image::TGA:
	default:
		PrintErrorMessage( "Format non pris en charge" );
	}

	return false;
}


// Sauvegarde une image dans un fichier
bool Image::Save( const char *filename, Image::Format format ) const
{
	switch( format )
	{
	case Image::PBM: return SavePNM( filename, PBM_BINARY );
	case Image::PGM: return SavePNM( filename, PGM_BINARY );
	case Image::PPM: return SavePNM( filename, PPM_BINARY );
	case Image::PNM: return SavePNM( filename );
	case Image::BMP: return SaveBMP( filename );
	case Image::JPG: return SaveJPG( filename );
	case Image::PNG: return SavePNG( filename );

	case Image::TGA:
	default:
		PrintErrorMessage( "Format non pris en charge" );
	}

	return false;
}


// Supprime proprement une image
void Image::Clear( void )
{
	if( m_pixels )
		delete [] m_pixels;

	m_pixels = 0;
}


// Permet de savoir si une image contient des donnes valides
bool Image::IsValid( void ) const
{
	return !!m_pixels;
}


// Permet de recuperer un pixel d'une image
unsigned char * Image::GetPixel( unsigned int i ) const
{
	assert( i < m_pixel_number );

	return &m_pixels[ i * m_chan_mult_depth ];
}
unsigned char * Image::GetPixel( unsigned int x, unsigned int y ) const
{
	assert( x < m_width && y < m_height );

	return &m_pixels[ (x + (m_width*y))* m_chan_mult_depth ];
}


// Accesseurs
unsigned int Image::GetWidth( void ) const
{
	return m_width;
}
unsigned int Image::GetHeight( void ) const
{
	return m_height;
}
unsigned char Image::GetChanNumber( void ) const
{
	return m_chan_number;
}


// Met  jour des donnes utiles
void Image::Update( void )
{
	m_pixel_number    = m_width * m_height;
	m_chan_mult_depth = m_chan_number * m_depth;
}
