// BMP

#include "bmp_util.h"




// Charge un fichier BMP
bool Image::LoadBMP( const char *filename )
{
	FILE *file = fopen( filename, "rb" );
	
	if( !file )
	{
		PrintErrorMessage( ERROR_UNABLE_OPEN_FILE, filename );
		return false;
	}

	bmp_file_header bmfh;
	fread( &bmfh, sizeof(bmp_file_header), 1, file );

	if( strncmp( (char *)bmfh.bfType, "BM", 2 ) )
	{
		PrintErrorMessage( ERROR_WRONG_TYPE, filename, "BMP" );
		fclose( file );
		return false;
	}

	bmp_info_header bmih;
	fread( &bmih, sizeof(bmp_info_header), 1, file );

	if( bmih.biCompression != 0 )
	{
		PrintErrorMessage( ERROR_WRONG_TYPE, filename, "BMP" );
		fclose( file );
		return false;
	}

	unsigned char *colormap = NULL;
	
	if( bmih.biBitCount <= 8 )
	{
		int colormapSize = (1 << bmih.biBitCount) * 4;
		colormap = (unsigned char *)malloc( colormapSize * sizeof(unsigned char) );

		if( !colormap )
		{
			PrintErrorMessage( ERROR_MEMORY );
			fclose( file );
			return false;
		}

		fread( colormap, sizeof(unsigned char), colormapSize, file );
	}

	fseek( file, bmfh.bfOffBits, SEEK_SET );

	unsigned char chan_number = 3;
	
	if( bmih.biBitCount == 1 )
		chan_number = 1;
	else if( bmih.biBitCount == 32 )
		chan_number = 4;
	
	unsigned char *pixels = (unsigned char *)malloc( bmih.biWidth * bmih.biHeight * chan_number * sizeof(unsigned char) );

	if( !pixels )
	{
		PrintErrorMessage( ERROR_MEMORY );
		if( colormap )
			free( colormap );
		fclose( file );
		return false;
	}

	switch( bmih.biBitCount )
	{
	case 1:
		for( int y = bmih.biHeight-1; y >= 0; y -- )
		{
			for( int x = 0; x < bmih.biWidth; )
			{
				unsigned char color = (unsigned char)fgetc( file );

				for( int j = 7; j >= 0 && x < bmih.biWidth; --j, x++ )
				{
					pixels[ x + bmih.biWidth*y ] = colormap[(((color & (1 << j)) > 0) * 4) ];
				}
			}
		}
		break;
	case 4:
		for( int y = bmih.biHeight-1; y >= 0; y -= 1 )
		{
			for( int x = 0; x < bmih.biWidth; x += 2 )
			{
				unsigned char color = (unsigned char)fgetc( file );
				unsigned char clrIndex = (color >> 4);
				pixels[ ((x + bmih.biWidth*y) * 3) + 2 ] = colormap[(clrIndex * 4) + 0];
				pixels[ ((x + bmih.biWidth*y) * 3) + 1 ] = colormap[(clrIndex * 4) + 1];
				pixels[ ((x + bmih.biWidth*y) * 3) + 0 ] = colormap[(clrIndex * 4) + 2];

				if( x+1 < bmih.biWidth )
				{
					clrIndex = (color & 0x0F);
					pixels[ ((x + bmih.biWidth*y) * 3) + 5 ] = colormap[(clrIndex * 4) + 0];
					pixels[ ((x + bmih.biWidth*y) * 3) + 4 ] = colormap[(clrIndex * 4) + 1];
					pixels[ ((x + bmih.biWidth*y) * 3) + 3 ] = colormap[(clrIndex * 4) + 2];
				}
			}
		}
		break;
	case 8:
		for( int y = bmih.biHeight-1; y >= 0; y-- )
		{
			for( int x = 0; x < bmih.biWidth; x++ )
			{
				unsigned char color = (unsigned char)fgetc( file );

				pixels[ ((x + bmih.biWidth*y) * 3) + 2 ] = colormap[(color * 4) + 0];
				pixels[ ((x + bmih.biWidth*y) * 3) + 1 ] = colormap[(color * 4) + 1];
				pixels[ ((x + bmih.biWidth*y) * 3) + 0 ] = colormap[(color * 4) + 2];
			}
		}
		break;
	case 24:
		for( int y = bmih.biHeight-1; y >= 0; y-- )
		{
			for( int x = 0; x < bmih.biWidth; x++ )
			{
				pixels[ ((x + bmih.biWidth*y) * 3) + 2 ] = (unsigned char)fgetc( file );
				pixels[ ((x + bmih.biWidth*y) * 3) + 1 ] = (unsigned char)fgetc( file );
				pixels[ ((x + bmih.biWidth*y) * 3) + 0 ] = (unsigned char)fgetc( file );
			}
		}
		break;
	case 32:
		for( int y = bmih.biHeight-1; y >= 0; y-- )
		{
			for( int x = 0; x < bmih.biWidth; x++ )
			{
				pixels[ ((x + bmih.biWidth*y) * 4) + 2 ] = (unsigned char)fgetc( file );
				pixels[ ((x + bmih.biWidth*y) * 4) + 1 ] = (unsigned char)fgetc( file );
				pixels[ ((x + bmih.biWidth*y) * 4) + 0 ] = (unsigned char)fgetc( file );
				pixels[ ((x + bmih.biWidth*y) * 4) + 3 ] = (unsigned char)fgetc( file );
			}
		}
		break;
	default:
		PrintErrorMessage( ERROR_UNKNOWN_TYPE, filename, "BMP" );
		if( colormap )
			free( colormap );
		free( pixels );
		fclose( file );
		return false;
		break;
	}

	if( colormap )
		free( colormap );

	fclose( file );

	Clear();

	m_width       = bmih.biWidth;
	m_height      = bmih.biHeight;
	m_pixels      = pixels;
	m_chan_number = chan_number;
	m_depth       = 1;

	Update();

	return true;
}




// Sauvegarde une image dans un fichier BMP
bool Image::SaveBMP( const char *filename ) const
{
	if( !IsValid() )
		return false;

	if( m_chan_number < 1 || m_chan_number > 4 )
	{
		PrintErrorMessage( ERROR_WRONG_CHAN_NUMBER, filename, "BMP" );
		return false;
	}

	FILE *file = fopen( filename, "wb" );
	if( !file )
	{
		PrintErrorMessage( ERROR_UNABLE_CREATE_FILE, filename );
		return false;
	}

	fclose( file );
	return true;
}
