// Joint Photographic Experts Group

#include "jpg_util.h"




// Charge un fichier JPG
bool Image::LoadJPG( const char *filename )
{
	FILE *file = fopen( filename, "rb" );

	if( !file )
	{
		PrintErrorMessage( ERROR_UNABLE_OPEN_FILE, filename );
		return false;
	}

	jpeg_decompress_struct cinfo;
	jpeg_error_struct      cerror;

	jpeg_create_decompress( &cinfo );
	cinfo.err = jpeg_std_error( &cerror.err );

	cerror.err.error_exit     = jpeg_error;
	cerror.err.output_message = jpeg_output;

	if( setjmp( cerror.jmp ) )
	{
		PrintErrorMessage( ERROR_CORRUPTED_FILE, filename );
		jpeg_destroy_decompress( &cinfo );
		fclose( file );
		return false;
	}

	jpeg_create_decompress( &cinfo );
	jpeg_stdio_src( &cinfo, file );

	jpeg_read_header( &cinfo, true );
	jpeg_start_decompress( &cinfo );

	unsigned int   width       = cinfo.image_width;
	unsigned int   height      = cinfo.image_height;
	unsigned char  chan_number = (unsigned char)cinfo.num_components;
	unsigned char *pixels      = new unsigned char[ width*height*chan_number ];

	if( !pixels )
	{
		PrintErrorMessage( ERROR_MEMORY );
		jpeg_destroy_decompress( &cinfo );
		fclose( file );
		return false;
	}

	JSAMPROW row;
	unsigned int row_stride = width*chan_number;

	for( unsigned int i = 0; i < height; ++i )
	{
		row = pixels + ( i*row_stride );
		jpeg_read_scanlines( &cinfo, &row, 1 );
	}

	jpeg_finish_decompress(  &cinfo );
	jpeg_destroy_decompress( &cinfo );
	fclose( file );

	Clear();

	m_width       = width;
	m_height      = height;
	m_pixels      = pixels;
	m_chan_number = chan_number;
	m_depth       = 1;

	Update();

	return true;
}




// Sauvegarde une image dans un fichier JPG
bool Image::SaveJPG( const char *filename, int quality ) const
{
	if( !IsValid() )
		return false;

	if( m_chan_number != 1 && m_chan_number != 3 )
	{
		PrintErrorMessage( ERROR_WRONG_CHAN_NUMBER, filename, "JPG" );
		return false;
	}

	FILE *file = fopen( filename, "wb" );

	if( !file )
	{
		PrintErrorMessage( ERROR_UNABLE_CREATE_FILE, filename );
		return false;
	}

	if( quality <   1 ) quality =   1;
	if( quality > 100 ) quality = 100;


	jpeg_compress_struct cinfo;
	jpeg_error_struct    cerror;

	cinfo.err = jpeg_std_error( &cerror.err );
	cerror.err.error_exit     = jpeg_error;
	cerror.err.output_message = jpeg_output;

	if( setjmp( cerror.jmp ) )
	{
		PrintErrorMessage( ERROR_UNABLE_SAVE_FILE, filename, "JPG" );
		jpeg_destroy_compress( &cinfo );
		fclose( file );
		return false;
	}

	jpeg_create_compress( &cinfo );
	jpeg_stdio_dest( &cinfo, file );

	cinfo.image_width      = m_width;
	cinfo.image_height     = m_height;
	cinfo.input_components = m_chan_number;
	cinfo.in_color_space   = (m_chan_number == 1) ? JCS_GRAYSCALE : JCS_RGB;
	jpeg_set_defaults( &cinfo );
	jpeg_set_quality(  &cinfo, quality, true );

	jpeg_start_compress( &cinfo, true );

	JSAMPROW row;
	unsigned int row_stride = m_width*m_chan_number*m_depth;

	while( cinfo.next_scanline < cinfo.image_height )
	{
		row = &m_pixels[ cinfo.next_scanline * row_stride ];
		jpeg_write_scanlines( &cinfo, &row, 1 );
	}

	jpeg_finish_compress(  &cinfo );
	jpeg_destroy_compress( &cinfo );
	fclose( file );

	return true;
}
