// Portable Newtork Graphics

#include "png_util.h"




// Charge un fichier PNG
bool Image::LoadPNG( const char *filename )
{
	FILE *file = fopen( filename, "rb" );
	
	if( !file )
	{
		PrintErrorMessage( ERROR_UNABLE_OPEN_FILE, filename );
		return false;
	}

	png_byte magic[8];
	fread( magic, 1, sizeof(magic), file );

	if( !png_check_sig( magic, sizeof(magic) ) )
	{
		PrintErrorMessage( ERROR_WRONG_TYPE, filename, "PNG" );
		fclose( file );
		return false;
	}

	png_structp png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING, NULL, png_error_fn, png_warning_fn );
	if( !png_ptr )
	{
		PrintErrorMessage( ERROR_MEMORY );
		fclose( file );
		return false;
	}

	png_infop info_ptr = png_create_info_struct( png_ptr );
	if( !info_ptr )
	{
		PrintErrorMessage( ERROR_MEMORY );
		png_destroy_read_struct( &png_ptr, NULL, NULL );
		fclose( file );
		return false;
	}

	if( setjmp( png_jmpbuf(png_ptr) ) )
	{
		PrintErrorMessage( ERROR_CORRUPTED_FILE, filename );
		png_destroy_read_struct( &png_ptr, &info_ptr, NULL );
		fclose( file );
		return false;
	}

	png_init_io( png_ptr, file );

	png_set_sig_bytes( png_ptr, sizeof( magic));

	png_read_info( png_ptr, info_ptr );

	int depth      = png_get_bit_depth(  png_ptr, info_ptr );
	int color_type = png_get_color_type( png_ptr, info_ptr );

	if( color_type == PNG_COLOR_TYPE_PALETTE)
		png_set_palette_to_rgb( png_ptr );

	if( color_type == PNG_COLOR_TYPE_GRAY && depth < 8 )
		png_set_gray_1_2_4_to_8( png_ptr );

	if( png_get_valid( png_ptr, info_ptr, PNG_INFO_tRNS ) )
		png_set_tRNS_to_alpha( png_ptr );

	if( depth == 16 )
		png_set_strip_16( png_ptr );
	else if( depth < 8 )
		png_set_packing( png_ptr );

	png_read_update_info( png_ptr, info_ptr );

	png_uint_32 width, height;
	png_get_IHDR( png_ptr, info_ptr, &width, &height, &depth, &color_type, NULL, NULL, NULL );

	unsigned int chan_number = 0;
	switch( color_type )
	{
	case PNG_COLOR_TYPE_GRAY      : chan_number = 1; break;
	case PNG_COLOR_TYPE_GRAY_ALPHA: chan_number = 2; break;
	case PNG_COLOR_TYPE_RGB       : chan_number = 3; break;
	case PNG_COLOR_TYPE_RGB_ALPHA : chan_number = 4; break;
	default:
		PrintErrorMessage( ERROR_CORRUPTED_FILE, filename );
		png_destroy_read_struct( &png_ptr, &info_ptr, NULL );
		fclose( file );
		return false;
	}

	unsigned char *pixels = new unsigned char[ width*height*chan_number ];
	if( !pixels )
	{
		PrintErrorMessage( ERROR_MEMORY );
		png_destroy_read_struct( &png_ptr, &info_ptr, NULL );
		fclose( file );
		return false;
	}

	png_bytep *rows = new png_bytep[ height ];
	if( !rows )
	{
		PrintErrorMessage( ERROR_MEMORY );
		png_destroy_read_struct( &png_ptr, &info_ptr, NULL );
		fclose( file );
		return false;
	}

	for( unsigned int i = 0; i < height; ++i )
		rows[i] = (png_bytep)( pixels + (i*width*chan_number) );

	png_read_image( png_ptr, rows );
	png_read_end(   png_ptr, NULL );
	png_destroy_read_struct( &png_ptr, &info_ptr, NULL );

	delete [] rows;

	fclose( file );

	Clear();

	m_width       = width;
	m_height      = height;
	m_pixels      = pixels;
	m_chan_number = chan_number;
	m_depth       = 1;

	Update();

	return true;
}




// Sauvegarde une image dans un fichier PNG
bool Image::SavePNG( const char *filename, FormatOption method ) const
{
	if( !IsValid() )
		return false;

	if( m_chan_number < 1 || m_chan_number > 4 )
	{
		PrintErrorMessage( ERROR_WRONG_CHAN_NUMBER, filename, "PNG" );
		return false;
	}

	FILE *file = fopen( filename, "wb" );
	if( !file )
	{
		PrintErrorMessage( ERROR_UNABLE_CREATE_FILE, filename );
		return false;
	}

	png_structp png_ptr = png_create_write_struct( PNG_LIBPNG_VER_STRING, (png_voidp)filename, png_error_fn, png_warning_fn );
	if( !png_ptr )
	{
		PrintErrorMessage( ERROR_MEMORY );
		fclose( file );
		return false;
	}

	png_infop info_ptr = png_create_info_struct( png_ptr );
	if( !info_ptr )
	{
		PrintErrorMessage( ERROR_MEMORY );
		png_destroy_write_struct( &png_ptr, NULL );
		fclose( file );
		return false;
	}

	if( setjmp( png_jmpbuf(png_ptr) ) )
	{
		PrintErrorMessage( ERROR_UNABLE_SAVE_FILE, filename, "PNG" );
		png_destroy_write_struct( &png_ptr, &info_ptr );
		fclose( file );
		return false;
	}

	png_init_io( png_ptr, file );

	int color_type = -1;
	switch( m_chan_number )
	{
	case 1: color_type = PNG_COLOR_TYPE_GRAY;       break;
	case 2: color_type = PNG_COLOR_TYPE_GRAY_ALPHA; break;
	case 3: color_type = PNG_COLOR_TYPE_RGB;        break;
	case 4: color_type = PNG_COLOR_TYPE_RGB_ALPHA;  break;
	}

	int interlace_method = (method == Image::NONE) ? PNG_INTERLACE_NONE : PNG_INTERLACE_ADAM7;
	png_set_IHDR( png_ptr, info_ptr, m_width, m_height, 8, color_type, interlace_method, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT );

	png_write_info( png_ptr, info_ptr );

	png_bytep *rows = new png_bytep[ m_height ];
	if( !rows )
	{
		PrintErrorMessage( ERROR_MEMORY );
		png_destroy_write_struct( &png_ptr, &info_ptr );
		fclose( file );
		return false;
	}

	for( unsigned int i = 0; i < m_height; ++i )
		rows[i] = (png_bytep)( m_pixels + (i*m_width*m_chan_number) );

	png_write_image( png_ptr, rows );
	png_write_end(   png_ptr, NULL );
	png_destroy_write_struct( &png_ptr, &info_ptr );

	delete [] rows;

	fclose( file );
	return true;
}
