// Portable AnyMap

#include "pnm_util.h"




// Charge un fichier PNM (PGM, PBM ou PPM)
bool Image::LoadPNM( const char *filename )
{
	FileBuffer file_buffer( filename );

	if( !file_buffer.IsValid() )
		return false;

	unsigned char *buffer = file_buffer.m_data;

	// Magic number P1 -> P6
	if( file_buffer.m_length < 8 || ( *buffer != 'P' && *buffer != 'p' ))
	{
		PrintErrorMessage( ERROR_WRONG_TYPE, filename, "PNM" );
		return false;
	}

	int format = Image::NONE;
	switch( *(++buffer) )
	{
	case '1': format = Image::PBM_ASCII;  break;
	case '2': format = Image::PGM_ASCII;  break;
	case '3': format = Image::PPM_ASCII;  break;
	case '4': format = Image::PBM_BINARY; break;
	case '5': format = Image::PGM_BINARY; break;
	case '6': format = Image::PPM_BINARY; break;

	default:
		PrintErrorMessage( ERROR_WRONG_TYPE, filename, "PNM" );
		return false;
	}

	int chan_number = 0;
	if( format == Image::PPM_ASCII || format == Image::PPM_BINARY )
		chan_number = 3;
	else
		chan_number = 1;

	// hauteur et largeur de l'image
	buffer     = JumpToNextChar( ++buffer );
	int width  = atoi( (const char *)buffer );
	buffer     = JumpToNextChar( ++buffer );
	int height = atoi( (const char *)buffer );

	if( width <= 0x0000 || width > 0xffff || height <= 0x0000 || height > 0xffff )
	{
		PrintErrorMessage( ERROR_WRONG_DIMENSIONS, filename );
		return false;
	}

	// Profondeur de l'image ( formats PGM et PPM uniquement )
	int max_value = 1;
	if( format != Image::PBM_ASCII && format != Image::PBM_BINARY )
	{
		buffer = JumpToNextChar( ++buffer );
		max_value = atoi( (const char*)buffer );
	}

	if( max_value <= 0 || max_value > 255 )
	{
		PrintErrorMessage( ERROR_WRONG_DEPTH, filename );
		return false;
	}
	
	// Allocation du tableau de pixels
	unsigned int value_number = width*height*chan_number;

	unsigned char *pixels = new unsigned char[ value_number ];
	if( !pixels )
	{
		PrintErrorMessage( ERROR_MEMORY );
		return false;
	}
	
	buffer = JumpToNextChar( ++buffer );

	// Format PGM et PPM
	if( format != Image::PBM_ASCII && format != Image::PBM_BINARY )
	{
		// ASCII
		if( format == Image::PGM_ASCII || format == Image::PPM_ASCII )
		{
			for( unsigned int i = 0; i < value_number; ++i )
			{
				if( !(*buffer) )
				{
					PrintErrorMessage( ERROR_CORRUPTED_FILE, filename );
					delete [] pixels;
					return false;
				}

				pixels[ i ] = (unsigned char)( (atoi((char *)buffer)*255)/max_value );

				buffer = JumpToNextChar( ++buffer );
			}
		}
		// Binaire
		else
		{
			if( &file_buffer.m_data[ file_buffer.m_length ] < &buffer[ value_number ] )
			{
				PrintErrorMessage( ERROR_CORRUPTED_FILE, filename );
				delete [] pixels;
				return false;
			}

			for( unsigned int i = 0; i < value_number; ++i )
				pixels[ i ] = (unsigned char)( ((int)buffer[i]*255)/max_value );
		}
	}
	// Format PBM
	else
	{
		// ASCII
		if( format == Image::PBM_ASCII )
		{
			for( unsigned int i = 0; i < value_number; ++i )
			{
				if( *buffer == 0 )
				{
					PrintErrorMessage( ERROR_CORRUPTED_FILE, filename );
					delete [] pixels;
					return false;
				}

				pixels[ i ] = (*buffer == '0')?255:0;
				buffer = JumpBlanks( ++buffer );
			}
		}
		// Binaire
		else
		{
			if( &file_buffer.m_data[ file_buffer.m_length ] < &buffer[ (unsigned int)(ceil((float)width / 8.0f))*height ] )
			{
				PrintErrorMessage( ERROR_CORRUPTED_FILE, filename );
				delete [] pixels;
				return false;
			}

			for( int row = 0; row < height; ++row )
			{
				int pix = 0;
				while( pix < width )
				{
					for( unsigned char mask = 128; (mask > 0) && (pix < width); mask >>= 1, ++pix )
						pixels[ pix + width*row ] = (*buffer & mask)?0:255;
					
					++buffer;
				}
			}
		}
	}

	Clear();

	m_width       = width;
	m_height      = height;
	m_pixels      = pixels;
	m_chan_number = chan_number;
	m_depth       = 1;

	Update();

	return true;
}




// Sauvegarde une image dans un fichier PNM
bool Image::SavePNM( const char *filename, FormatOption option ) const
{
	if( !IsValid() )
		return false;

	if( option < PBM_ASCII || option > PPM_BINARY )
	{
		option = (m_chan_number > 2) ? Image::PPM_BINARY : Image::PGM_BINARY;
	}

#define LOG10(x) ((int)(log((float)x)/2.30259f)+1)
	unsigned int buffer_size = 2;
	++buffer_size += LOG10(m_width );
	++buffer_size += LOG10(m_height);
#undef LOG10

	if( option != Image::PBM_ASCII && option != Image::PBM_BINARY )
		++buffer_size += 3;

	unsigned int chan_number = 0;
	switch( option )
	{
	case Image::PBM_ASCII : chan_number = 1; break;
	case Image::PGM_ASCII : chan_number = 1; break;
	case Image::PPM_ASCII : chan_number = 3; break;
	case Image::PBM_BINARY: chan_number = 1; break;
	case Image::PGM_BINARY: chan_number = 1; break;
	case Image::PPM_BINARY: chan_number = 3; break;

	default:
		PrintErrorMessage( "tamaire" );
		return false;
	}

	bool binary_mode = (option == Image::PBM_BINARY || option == Image::PGM_BINARY || option == Image::PPM_BINARY);


	if( option == Image::PBM_BINARY )
		++buffer_size += (unsigned int)(ceil((float)m_width / 8.0f))*m_height;
	else if( option == Image::PBM_ASCII )
		++buffer_size += m_pixel_number*2;
	else if( binary_mode )
		++buffer_size += m_pixel_number*chan_number;
	else
		++buffer_size += m_pixel_number*chan_number*(3+1);

	unsigned char *buffer = new unsigned char[ buffer_size ];

	if( !buffer )
	{
		PrintErrorMessage( ERROR_MEMORY );
		return false;
	}

	unsigned char *p = 0;
	switch( option )
	{
	case Image::PBM_ASCII : p =  &buffer[ sprintf( (char *)buffer, "P1\n%u %u",        m_width, m_height )-1 ]; break;
	case Image::PGM_ASCII : buffer_size = sprintf( (char *)buffer, "P2\n%u %u\n255\n", m_width, m_height );     break;
	case Image::PPM_ASCII : buffer_size = sprintf( (char *)buffer, "P3\n%u %u\n255\n", m_width, m_height );     break;
	case Image::PBM_BINARY: p =  &buffer[ sprintf( (char *)buffer, "P4\n%u %u\n"     , m_width, m_height )-1 ]; break;
	case Image::PGM_BINARY: p =  &buffer[ sprintf( (char *)buffer, "P5\n%u %u\n255\n", m_width, m_height )-1 ]; break;
	case Image::PPM_BINARY: p =  &buffer[ sprintf( (char *)buffer, "P6\n%u %u\n255\n", m_width, m_height )-1 ]; break;
	default: PrintErrorMessage( "Internal error" ); break;
	}

	if( option != Image::PBM_ASCII && option != Image::PBM_BINARY )
	{
		// PGM et PPM binaires
		if( binary_mode )
		{
			unsigned int i, j;
			unsigned char *temp;
			int g;

			if( chan_number == 1 && m_chan_number < 3 )
			{
				for( i = 0; i < m_pixel_number; ++i )
					*(++p) = m_pixels[ i * m_chan_mult_depth ];
			}
			else if( chan_number == 1 && m_chan_number > 2 )
			{
				for( i = 0; i < m_pixel_number; ++i )
				{
					temp = &m_pixels[ i * m_chan_mult_depth ];
					g = (temp[0]+temp[1]+temp[2])/3;
					*(++p) = (unsigned char)g;
				}
			}
			else if( chan_number == 3 && m_chan_number < 3 )
			{
				for( i = 0; i < m_pixel_number; ++i )
				{
					temp = &m_pixels[ i * m_chan_mult_depth ];
					for( j = 0; j < 3; ++j )
						*(++p) = temp[0];
				}
			}
			else //if( chan_number == 3 && m_chan_number > 2 )
			{
				for( i = 0; i < m_pixel_number; ++i )
				{
					temp = &m_pixels[ i * m_chan_mult_depth ];
					for( j = 0; j < 3; ++j )
						*(++p) = temp[j];
				}
			}
		}
		// PGM et PPM ASCII
		else
		{
			unsigned int jump = (m_width < 70)?m_width:70;
			unsigned char *pix;

			if( chan_number == 1 && m_chan_number < 3 )
			{
				for( unsigned int i = 0; i < m_pixel_number; ++i )
					buffer_size += sprintf( (char *)&buffer[ buffer_size ], "%c%d", ((i%jump)?' ':'\n'), m_pixels[ i * m_chan_mult_depth ] );
			}
			else if( chan_number == 1 && m_chan_number > 2 )
			{
				int g;
				for( unsigned int i = 0; i < m_pixel_number; ++i )
				{
					pix = &m_pixels[ i * m_chan_mult_depth ];
					g = (pix[0]+pix[1]+pix[2])/3;
					buffer_size += sprintf( (char *)&buffer[ buffer_size ], "%c%d", ((i%jump)?' ':'\n'), g );
				}
			}
			else if( chan_number == 3 && m_chan_number < 3 )
			{
				for( unsigned int i = 0; i < m_pixel_number; ++i )
				{
					pix = &m_pixels[ i * m_chan_mult_depth ];
					buffer_size += sprintf( (char *)&buffer[ buffer_size ], "%c%d %d %d", ((i%jump)?' ':'\n'), pix[0],pix[0],pix[0] );
				}
			}
			else //if( chan_number == 3 && m_chan_number > 2 )
			{
				for( unsigned int i = 0; i < m_pixel_number; ++i )
				{
					pix = &m_pixels[ i * m_chan_mult_depth ];
					buffer_size += sprintf( (char *)&buffer[ buffer_size ], "%c%d %d %d", ((i%jump)?' ':'\n'), pix[0],pix[1],pix[2] );
				}
			}
		}
	}
	else
	{
		// PBM binaire
		if( binary_mode )
		{
			unsigned int row, pix;
			unsigned char c, mask;
			if( m_chan_number > 2 )
			{
				unsigned char *temp;
				int g;

				for( row = 0; row < m_height; ++row )
				{
					for( pix = 0; pix < m_width; )
					{
						for( c = 0, mask = 128; (mask > 0) && (pix < m_width); mask >>= 1, ++pix )
						{
							temp = &m_pixels[ (pix + (m_width*row))* m_chan_mult_depth ];
							g = (temp[0]+temp[1]+temp[2]);

							if( g < 384 )
								c += mask;
						}
						*(++p) = c;
					}
				}
			}
			else
			{
				for( row = 0; row < m_height; ++row )
				{
					for( pix = 0; pix < m_width; )
					{
						for( c = 0, mask = 128; (mask > 0) && (pix < m_width); mask >>= 1, ++pix )
						{
							if( m_pixels[ (pix + (m_width*row))* m_chan_mult_depth ] < 128 )
								c += mask;
						}
						*(++p) = c;
					}
				}
			}
		}
		// PBM ASCII
		else
		{
			--buffer_size;

			unsigned int i, jump = (m_width < 70)?m_width:70;

			if( m_chan_number > 2 )
			{
				unsigned char *temp;
				int g;
				for( i = 0; i < m_pixel_number; ++i )
				{
					temp = &m_pixels[ i * m_chan_mult_depth ];
					g = (temp[0]+temp[1]+temp[2]);
					*(++p) = (i%jump)?' ':'\n';
					*(++p) = (g < 384)?'1':'0';
				}
			}
			else
			{
				for( i = 0; i < m_pixel_number; ++i )
				{
					*(++p) = (i%jump)?' ':'\n';
					*(++p) = (m_pixels[ i ] < 128)?'1':'0';
				}
			}
		}
	}

	FileBuffer file_buffer( buffer, buffer_size );

	return file_buffer.WriteFile( filename );
}
