#include "fourier.h"
#include <cassert>
#include <cmath>
#include <float.h>
#ifndef M_PI
#define M_PI 3.14159265358979323846f
#endif


unsigned int nb_bits_utile(unsigned long v)
{
	unsigned int n = 0;
	while( v > 0 )
	{
		v >>= 1;
		n++;
	}
	return n;
}

Fourier::Fourier( const Image& image )
{
	m_image.Copy( &image );

	m_C = image.GetWidth();
	m_L = image.GetHeight();
	
	m_X = new Complex*[ m_L ];
	m_XTemp = new Complex*[ m_L ];
	m_XSave = new Complex*[ m_L ];
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		m_X[i] = new Complex[ m_C ];
		m_XTemp[i] = new Complex[ m_C ];
		m_XSave[i] = new Complex[ m_C ];
		
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] = Complex( (double)image.GetPixel( j, i )[0], 0.0f );
		}
	}
	m_Exp = new Complex[ m_L ];

	FastTransform();
}


Fourier::~Fourier( void )
{
	for( unsigned int i = 0 ; i < m_L ; i++ )
	{
		delete [] m_XTemp[i];
		delete [] m_X[i];
	}
	delete [] m_XTemp;
	delete [] m_X;
}


void Fourier::Transform()
{
	double LC_inv = 1.0 / (double)(m_L * m_C);
	
	for( unsigned int u = 0; u < m_L; u++ )
	{
		for( unsigned int v = 0; v < m_C; v++ )
		{
			Complex sum_l( 0.0, 0.0 );
			
			for( unsigned int l = 0; l < m_L; l++ )
			{
				Complex sum_c( 0.0, 0.0 );
				
				for( unsigned int c = 0; c < m_C; c++ )
				{
					unsigned char* x = m_image.GetPixel( c, l );
					
					sum_c += static_cast< double >( x[0] ) * std::exp( Complex( 0.0, -2.0 * M_PI * ( ((u*l) / (double)m_L) + ((v*c) / (double)m_C) ) ) );
				}
				
				sum_l += sum_c;
			}
			
			m_X[ u ][ v ] = LC_inv * sum_l;
		}
	}
}


Image Fourier::TransformInverse( void )
{
	assert( m_X );
	
	Image image_result( m_C, m_L, 1 );
	
	for( unsigned int l = 0; l < m_L; l++ )
	{
		for( unsigned int c = 0; c < m_C; c++ )
		{
			Complex sum_u( 0.0, 0.0 );
			
			for( unsigned int u = 0; u < m_L; u++ )
			{
				Complex sum_v( 0.0, 0.0 );
				
				for( unsigned int v = 0; v < m_C; v++ )
				{
					sum_v += m_X[ u ][ v ] * std::exp( Complex( 0.0, 2.0 * M_PI * ( ((u*l) / (double)m_L) + ((v*c) / (double)m_C) ) ) );
				}
				
				sum_u += sum_v;
			}
			
			image_result.GetPixel( c, l )[0] = (unsigned char)( sum_u.real() + 0.5 );
		}
	}
	
	return image_result;
}

void Fourier::FastTransform()
{
	ComputeExp();

	Reorganisation();
	
	unsigned int i, j;
	unsigned int s = 2;
	while( s <= m_C )
	{
		for( i = 0 ; i < m_L ; i += s )
		{
			for( j = 0 ; j < m_C ; j += s )
			{
				Butterfly( i, j, s, -1 );
			}
		}
			
		s <<= 1;
	}
	
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] /= m_L*m_C;
		}
	}
	//Centre la reprsentation
	Center();
}

Image Fourier::FastTransformInverse( void )
{
	//Dcentre la reprsentation
	Center();
	Reorganisation();
	
	unsigned int s = 2;
	while( s <= m_C )
	{
		for( unsigned int i = 0 ; i < m_L ; i += s )
		{
			for( unsigned int j = 0 ; j < m_C ; j += s )
			{
				Butterfly( i, j, s, 1 );
			}
		}
			
		s <<= 1;
	}
	
	Image image_result( m_C, m_L, 1 );
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			int c = (int)( m_X[i][j].real() + 0.5 );
			if( c < 0 ) c = 0;
			if( c > 255 ) c = 255;
			image_result.GetPixel( j, i )[0] = c;
		}
	}
	return image_result;
}

void Fourier::InverseBits( unsigned int* value, unsigned int nb_bits )
{
	unsigned int v = *value;
	unsigned int result = 0;
	while( --nb_bits )
	{
		result |= v & 1;
		result <<= 1;
		v >>= 1;
	}
	*value = result + (v & 1);
}

void Fourier::Reorganisation()
{
	unsigned int i, j, i_inv, j_inv;	
	unsigned int log_size;
	log_size = nb_bits_utile( m_image.GetHeight() ) - 1;
	
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			i_inv = i;
			InverseBits( &i_inv, log_size );
			j_inv = j;
			InverseBits( &j_inv, log_size );

			m_XTemp[i][j] = m_X[i_inv][j_inv];
		}
	}
	
	for( i = 0 ; i < m_L ; ++i )
	{
		for( j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] = m_XTemp[i][j];
		}
	}
}

void Fourier::ComputeExp()
{
	for( unsigned int i = 0 ; i < m_L ; i++ )
	{
		m_Exp[i] = std::exp( Complex( 0.0, M_PI * i / (double)(m_L/2) ) );
	}
}

void Fourier::Butterfly( const unsigned int l, const unsigned int c, const unsigned int size, const int sign )
{
	Complex Xpp, Xip, Xpi, Xii;
	Complex e_u, e_v;

	for( unsigned int i = l ; i < l+size/2 ; i++ )
	{
		e_u = Complex( m_Exp[(i*m_L/size)%m_L].real(), sign *  m_Exp[(i*m_L/size)%m_L].imag() );
		for( unsigned int j = c ; j < c+size/2 ; j++ )
		{
			e_v = Complex( m_Exp[(j*m_C/size)%m_C].real(), sign * m_Exp[(j*m_C/size)%m_C].imag() );
			
			Xpp =             m_X[i       ][j       ];
			Xip = e_u *       m_X[i+size/2][j       ];
			Xpi =       e_v * m_X[i       ][j+size/2];
			Xii = e_u * e_v * m_X[i+size/2][j+size/2];
			
			Complex Xgplus  = Xpp + Xip;
			Complex Xgmoins = Xpp - Xip;
			Complex Xdplus  = Xpi + Xii;
			Complex Xdmoins = Xpi - Xii;

			m_X[i       ][j       ] = Xgplus  + Xdplus;
			m_X[i+size/2][j       ] = Xgmoins + Xdmoins;
			m_X[i       ][j+size/2] = Xgplus  - Xdplus;
			m_X[i+size/2][j+size/2] = Xgmoins - Xdmoins;
			
		}
	}
}

void Fourier::Center()
{
	unsigned int i, j, i_c, j_c;
	unsigned int log_size;
	log_size = nb_bits_utile( m_image.GetHeight() ) - 1;
	
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			i_c = i;
			j_c = j;

			if( i <  m_L/2 && j <  m_C/2 ) { i_c = i + m_L/2; j_c = j + m_C/2; }
			if( i >= m_L/2 && j <  m_C/2 ) { i_c = i - m_L/2; j_c = j + m_C/2; }
			if( i <  m_L/2 && j >= m_C/2 ) { i_c = i + m_L/2; j_c = j - m_C/2; }
			if( i >= m_L/2 && j >= m_C/2 ) { i_c = i - m_L/2; j_c = j - m_C/2; }
			
			m_XTemp[i_c][j_c] = m_X[i][j];
		}
	}
	
	for( i = 0 ; i < m_L ; ++i )
	{
		for( j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] = m_XTemp[i][j];
		}
	}
}

Image Fourier::GetResult()
{
	double** m_temp = new double*[ m_L ];
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		m_temp[i] = new double[ m_C ];
	}
	
	double min = FLT_MAX, max = -FLT_MAX;
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			m_temp[i][j] = std::norm( m_X[i][j] );
			m_temp[i][j] = std::log( std::log( std::log( std::log( std::log( std::log( std::log( m_temp[i][j]+1.0 )+1.0 ) +1.0) +1.0) +1.0) +1.0) +1.0);
			
			if( m_temp[i][j] < min ) min = m_temp[i][j];
			if( m_temp[i][j] > max ) max = m_temp[i][j];
		}
	}
	
	Image image_result( m_C, m_L, 1 );
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			image_result.GetPixel( j, i )[0] = (unsigned char)( ((m_temp[i][j]-min)*255.0) / (max-min) + 0.5 );
		}
	}
	
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		delete[] m_temp[i];
	}
	delete[] m_temp;
	
	return image_result;
}

void Fourier::IdealFilter( unsigned int radius )
{
	unsigned int i, j;
	unsigned int log_size;
	log_size = nb_bits_utile( m_image.GetHeight() ) - 1;
	
	for( i = 0 ; i < m_L ; ++i )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			m_XTemp[i][j] = Complex( 0.0, 0.0 );
		}
	}


	int f = 1 - radius;
	int ddF_x = 0;
	int ddF_y = -2 * radius;
	int x = 0;
	int y = radius;
 
	for( j = m_L/2 - radius ; j <= m_L/2 + radius ; j++ ) m_XTemp[m_C/2][j] = m_X[m_C/2][j];
	
	while(x < y) 
	{
		if(f >= 0) 
		{
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x + 1;
		
		for( j = m_L/2 - x ; j <= m_L/2 + x ; j++ ) m_XTemp[m_C/2+y][j] = m_X[m_C/2+y][j];
		for( j = m_L/2 - x ; j <= m_L/2 + x ; j++ ) m_XTemp[m_C/2-y][j] = m_X[m_C/2-y][j];
		for( j = m_L/2 - y ; j <= m_L/2 + y ; j++ ) m_XTemp[m_C/2+x][j] = m_X[m_C/2+x][j];
		for( j = m_L/2 - y ; j <= m_L/2 + y ; j++ ) m_XTemp[m_C/2-x][j] = m_X[m_C/2-x][j];
	}
	
	for( i = 0 ; i < m_L ; ++i )
	{
		for( j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] = m_XTemp[i][j];
		}
	}
}

void Fourier::GaussianFilter( double sigma )
{
	unsigned int i, j;
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			m_X[i][j] *= std::exp( Complex( -(sigma*sigma) * ( (i-m_L/2)*(i-m_L/2) + (j-m_C/2)*(j-m_C/2) ) / 2.0, 0.0 ) );
		}
	}
}

void Fourier::Save()
{
	unsigned int i, j;
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			m_XSave[i][j] = m_X[i][j];
		}
	}
}
void Fourier::Load()
{
	unsigned int i, j;
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			 m_X[i][j] = m_XSave[i][j];
		}
	}
}
