#include "fourierMT.h"
#include <cassert>
#include <cmath>
#include <float.h>
#ifndef M_PI
#define M_PI 3.14159265358979323846f
#endif

FourierMT::FourierMT( const Image& image )
{
	m_image.Copy( &image );

	m_C = image.GetWidth();
	m_L = image.GetHeight();
	
	m_X = new Complex*[ m_L ];
	m_XTemp = new Complex*[ m_L ];
	m_XSave = new Complex*[ m_L ];
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		m_X[i] = new Complex[ m_C ];
		m_XTemp[i] = new Complex[ m_C ];
		m_XSave[i] = new Complex[ m_C ];
		
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] = Complex( (double)image.GetPixel( j, i )[0], 0.0f );
		}
	}
	m_Exp = new Complex[ m_L ];

	m_threads_params = new ThreadParams[ 4 ];
	m_threads = new SDL_Thread* [ 4 ];
      
	for( int i = 0; i < 4; i++ )
	{
		m_threads_params[i].X = m_X;
		m_threads_params[i].Exp   = m_Exp;
		m_threads_params[i].size = m_L;
	}
	m_threads_params[0].x = 0;
	m_threads_params[0].y = 0;
	m_threads_params[1].x = m_L / 2;
	m_threads_params[1].y = 0;
	m_threads_params[2].x = 0;
	m_threads_params[2].y = m_L / 2;
	m_threads_params[3].x = m_L / 2;
	m_threads_params[3].y = m_L / 2;

	FastTransform();
}


FourierMT::~FourierMT( void )
{
	for( unsigned int i = 0 ; i < m_L ; i++ )
	{
		delete [] m_XTemp[i];
		delete [] m_X[i];
	}
	delete [] m_XTemp;
	delete [] m_X;

	delete [] m_threads;
	delete [] m_threads_params;
}



void FourierMT::FastTransform()
{
	ComputeExp();

	Reorganisation();
	
	for( int i = 0; i < 4; i++ )
	{
		m_threads_params[i].sign = -1;
		m_threads[i] = SDL_CreateThread( FourierMT::MTTransform, &( m_threads_params[i] ) );
	}
	for( int i = 0; i < 4; i++ )
	{
		SDL_WaitThread( m_threads[i], NULL );
	}

	
	Butterfly( 0, 0, m_C, -1, m_X, m_Exp, m_C );


	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] /= m_L*m_C;
		}
	}
	

	//Centre la reprsentation
	Center();
}

Image FourierMT::FastTransformInverse( void )
{
	//Dcentre la reprsentation
	Center();
	Reorganisation();
	
	for( int i = 0; i < 4; i++ )
	{
		m_threads_params[i].sign = 1;
		m_threads[i] = SDL_CreateThread( FourierMT::MTTransform, &( m_threads_params[i] ) );
	}
	for( int i = 0; i < 4; i++ )
	{
		SDL_WaitThread( m_threads[i], NULL );
	}

	Butterfly( 0, 0, m_C, 1, m_X, m_Exp, m_C );
	
	Image image_result( m_C, m_L, 1 );
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			int c = (int) (m_X[i][j].real() + 0.5);
			if( c < 0 ) c = 0;
			if( c > 255 ) c = 255;
			image_result.GetPixel( j, i )[0] = c;
		}
	}
	return image_result;
}

unsigned int FourierMT::NbBits( unsigned long v )
{
	unsigned int n = 0;
	while( v > 0 )
	{
		v >>= 1;
		n++;
	}
	return n;
}

void FourierMT::InverseBits( unsigned int* value, unsigned int nb_bits )
{
	unsigned int v = *value;
	unsigned int result = 0;
	while( --nb_bits )
	{
		result |= v & 1;
		result <<= 1;
		v >>= 1;
	}
	*value = result | (v & 1);
}

void FourierMT::Reorganisation()
{
	unsigned int i, j, i_inv, j_inv;	
	unsigned int log_size;
	log_size = NbBits( m_image.GetHeight() ) - 1;
	
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			i_inv = i;
			InverseBits( &i_inv, log_size );
			j_inv = j;
			InverseBits( &j_inv, log_size );

			m_XTemp[i][j] = m_X[i_inv][j_inv];
		}
	}
	
	for( i = 0 ; i < m_L ; ++i )
	{
		for( j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] = m_XTemp[i][j];
		}
	}
}

void FourierMT::ComputeExp()
{
	for( unsigned int i = 0 ; i < m_L ; i++ )
	{
		m_Exp[i] = std::exp( Complex( 0.0, M_PI * i / (double)(m_L/2) ) );
	}
}

void FourierMT::Center()
{
	unsigned int i, j, i_c, j_c;
	unsigned int log_size;
	log_size = NbBits( m_image.GetHeight() ) - 1;
	
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			i_c = i;
			j_c = j;

			if( i <  m_L/2 && j <  m_C/2 ) { i_c = i + m_L/2; j_c = j + m_C/2; }
			if( i >= m_L/2 && j <  m_C/2 ) { i_c = i - m_L/2; j_c = j + m_C/2; }
			if( i <  m_L/2 && j >= m_C/2 ) { i_c = i + m_L/2; j_c = j - m_C/2; }
			if( i >= m_L/2 && j >= m_C/2 ) { i_c = i - m_L/2; j_c = j - m_C/2; }
			
			m_XTemp[i_c][j_c] = m_X[i][j];
		}
	}
	
	for( i = 0 ; i < m_L ; ++i )
	{
		for( j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] = m_XTemp[i][j];
		}
	}
}

int FourierMT::MTTransform( void* data )
{
	ThreadParams* params = static_cast< ThreadParams* >( data );

	unsigned int l, c;
	unsigned int s = 2;
	while( s <= params->size/2 )
	{
		for( l = params->y ; l < params->y+params->size/2 ; l += s )
		{
			for( c = params->x ; c < params->x+params->size/2 ; c += s )
			{
				FourierMT::Butterfly( l, c, s, params->sign, params->X, params->Exp, params->size );
			}
		}
			
		s <<= 1;
	}

	return 0;
}

void FourierMT::Butterfly( const unsigned int l, const unsigned int c, const unsigned int size, const int sign, Complex** X, Complex* Exp, const unsigned int image_size )
{
	for( unsigned int i = l ; i < l+size/2 ; i++ )
	{
		const Complex e_u = Complex( Exp[(i*image_size/size)%image_size].real(), sign *  Exp[(i*image_size/size)%image_size].imag() );
		for( unsigned int j = c ; j < c+size/2 ; j++ )
		{
			const Complex e_v = Complex( Exp[(j*image_size/size)%image_size].real(), sign * Exp[(j*image_size/size)%image_size].imag() );
			
			const Complex Xpp =             X[i       ][j       ];
			const Complex Xip = e_u *       X[i+size/2][j       ];
			const Complex Xpi =       e_v * X[i       ][j+size/2];
			const Complex Xii = e_u * e_v * X[i+size/2][j+size/2];
			
			const Complex Xgplus  = Xpp + Xip;
			const Complex Xgmoins = Xpp - Xip;
			const Complex Xdplus  = Xpi + Xii;
			const Complex Xdmoins = Xpi - Xii;

			X[i       ][j       ] = Xgplus  + Xdplus;
			X[i+size/2][j       ] = Xgmoins + Xdmoins;
			X[i       ][j+size/2] = Xgplus  - Xdplus;
			X[i+size/2][j+size/2] = Xgmoins - Xdmoins;
			
		}
	}
}



Image FourierMT::GetResult()
{
	double** m_temp = new double*[ m_L ];
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		m_temp[i] = new double[ m_C ];
	}
	
	double min = FLT_MAX, max = -FLT_MAX;
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			m_temp[i][j] = std::norm( m_X[i][j] );
			m_temp[i][j] = std::log( std::log( std::log( std::log( std::log( std::log( std::log( m_temp[i][j]+1.0 )+1.0 ) +1.0) +1.0) +1.0) +1.0) +1.0);
			
			if( m_temp[i][j] < min ) min = m_temp[i][j];
			if( m_temp[i][j] > max ) max = m_temp[i][j];
		}
	}
	
	Image image_result( m_C, m_L, 1 );
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		for( unsigned int j = 0 ; j < m_C ; ++j )
		{
			image_result.GetPixel( j, i )[0] = (unsigned char)( ((m_temp[i][j]-min)*255.0) / (max-min) + 0.5 );
		}
	}
	
	for( unsigned int i = 0 ; i < m_L ; ++i )
	{
		delete[] m_temp[i];
	}
	delete[] m_temp;
	
	return image_result;
}

void FourierMT::IdealFilter( unsigned int radius )
{
	unsigned int i, j;
	unsigned int log_size;
	log_size = NbBits( m_image.GetHeight() ) - 1;
	
	for( i = 0 ; i < m_L ; ++i )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			m_XTemp[i][j] = Complex( 0.0, 0.0 );
		}
	}


	int f = 1 - radius;
	int ddF_x = 0;
	int ddF_y = -2 * radius;
	int x = 0;
	int y = radius;
 
	for( j = m_L/2 - radius ; j <= m_L/2 + radius ; j++ ) m_XTemp[m_C/2][j] = m_X[m_C/2][j];
	
	while(x < y) 
	{
		if(f >= 0) 
		{
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x + 1;
		
		for( j = m_L/2 - x ; j <= m_L/2 + x ; j++ ) m_XTemp[m_C/2+y][j] = m_X[m_C/2+y][j];
		for( j = m_L/2 - x ; j <= m_L/2 + x ; j++ ) m_XTemp[m_C/2-y][j] = m_X[m_C/2-y][j];
		for( j = m_L/2 - y ; j <= m_L/2 + y ; j++ ) m_XTemp[m_C/2+x][j] = m_X[m_C/2+x][j];
		for( j = m_L/2 - y ; j <= m_L/2 + y ; j++ ) m_XTemp[m_C/2-x][j] = m_X[m_C/2-x][j];
	}
	
	for( i = 0 ; i < m_L ; ++i )
	{
		for( j = 0 ; j < m_C ; ++j )
		{
			m_X[i][j] = m_XTemp[i][j];
		}
	}
}

void FourierMT::GaussianFilter( double sigma )
{
	unsigned int i, j;
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			m_X[i][j] *= std::exp( Complex( -(sigma*sigma) * ( (i-m_L/2)*(i-m_L/2) + (j-m_C/2)*(j-m_C/2) ) / 2.0, 0.0 ) );
		}
	}
}

void FourierMT::Save()
{
	unsigned int i, j;
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			m_XSave[i][j] = m_X[i][j];
		}
	}
}
void FourierMT::Load()
{
	unsigned int i, j;
	for( i = 0 ; i < m_L ; i++ )
	{
		for( j = 0 ; j < m_C ; j++ )
		{
			 m_X[i][j] = m_XSave[i][j];
		}
	}
}
