#include "hough.h"
#include "hough_utils.h"


#include <cstdio>
#include <cstring>

#include <SDL/SDL_thread.h>




// Prototypes
void VotePoint( int x, int y, int* data, int width, int height );
void VoteCircle( int x0, int y0, int radius, int* data, int width, int height );
void DrawCircle( Image* im, int x0, int y0, int radius );
void DrawPoint( Image* im, int x, int y );


// Multi-threading
struct ThreadParams
{
	static int    space_width;
	static int    space_height;
	static int    space_offset;
	static Image* source;
	static int    r_min;

	int* space;
	int r;
};
int    ThreadParams::space_width  = 0;
int    ThreadParams::space_height = 0;
int    ThreadParams::space_offset = 0;
Image* ThreadParams::source       = NULL;
int    ThreadParams::r_min        = 0;

int ComputeCircleSpace( void* data )
{
	ThreadParams* params = static_cast< ThreadParams* >( data );

	const int space_width  = ThreadParams::space_width;
	const int space_height = ThreadParams::space_height;
	const int space_offset = ThreadParams::space_offset;
	int*      space        = params->space;
	const Image* source    = ThreadParams::source;
	const int r_min        = ThreadParams::r_min;
	const int r            = params->r;

	for( unsigned int y = 0; y < source->GetHeight(); y++ )
	{
		for( unsigned int x = 0; x < source->GetWidth(); x++ )
		{
			if( source->GetPixel( x, y )[ 0 ] > 127 )
			{
				const int x_c = x + space_offset;
				const int y_c = y + space_offset;

				VoteCircle( x_c, y_c, r_min + r*2, space, space_width, space_height );
			}
		}
	}

	return EXIT_SUCCESS;
}


// Detection des cercles
void Hough::GetCircles( Image* dest, int number, float tolerance )
{
	const int width  = m_source.GetWidth();
	const int height = m_source.GetHeight();
	const int pixel_number = width*height;

	PrepareDestination( dest, &m_source );

	// Initialisations
	const int r_min   = 2;
	const int r_max   = Min( width, height ) / 2;
	const int r_range = (r_max - r_min) / 2 + 1;

	printf( "Detection de cercles, %d rayons [%d,%d], tolerance: %d%%\n", r_range, r_min, r_max, (int)(tolerance*100.0) );

	const int spaces_offset = r_max;
	const int spaces_width  = width  + 2*spaces_offset;
	const int spaces_height = height + 2*spaces_offset;
	const int spaces_pixel_number = spaces_width * spaces_height;
	int** spaces;
	spaces = new int*[ r_range ];

	for( int i = 0; i < r_range; i++ )
	{
		spaces[i] = new int[ spaces_pixel_number ];
		memset( spaces[i], 0, sizeof(*spaces[i])*spaces_pixel_number );
	}

	// Calcul des votants
	ThreadParams::space_width  = spaces_width;
	ThreadParams::space_height = spaces_height;
	ThreadParams::space_offset = spaces_offset;
	ThreadParams::source       = &m_source;
	ThreadParams::r_min        = r_min;

	ThreadParams* threads_params = new ThreadParams[ r_range ];
	SDL_Thread**  threads        = new SDL_Thread* [ r_range ];
	for ( int i = 0; i < r_range; i++ )
	{
		threads_params[i].r = i;
		threads_params[i].space = spaces[i];

		threads[i] = SDL_CreateThread( ComputeCircleSpace, &( threads_params[i] ) );
	}

	for( int i = 0; i < r_range; i++ )
	{
		SDL_WaitThread( threads[i], NULL );
	}

	delete [] threads;
	delete [] threads_params;

	/*for( int r = 0; r < r_range; r++ )
	{
		char filename[ 255 ];
		sprintf( filename, "hough_cercle%03d.png", r+1 );

		Image image;
		image.Create( spaces_width, spaces_height, 1 );
		SaveToImage( &image, spaces[r], spaces_width, spaces_height );
		image.SavePNG( filename );
	}*/

	// Extraction des cercles
	while( number > 0 )
	{
		int votes = -1;
		int c_r;
		int c_x;
		int c_y;

		// On recupere le max
		for( int r = 0; r < r_range; r++ )
		{
			const int radius = r_min + r*2;
			const int perimeter = (int)(2.0*M_PI*radius);

			for( int y = 0; y < height; y++ )
			{
				for( int x = 0; x < width; x++ )
				{
					const int index = (spaces_offset + x) + spaces_width*(y + spaces_offset);

					const float ratio = spaces[ r ][ index ]/(float)perimeter;

					if( ratio < tolerance )
						continue;

					if( votes < spaces[ r ][ index ] )
					{
						votes = spaces[ r ][ index ];
						c_r = r;
						c_x = x;
						c_y = y;
					}
				}
			}
		}

		// On dessine le cercle s'il y en a un
		if( votes < 8 )
		{
			break;
		}

		// On supprime ce cercle de l'espace de Hough
		const int radius = r_min + c_r*2;
		const int perimeter = (int)(2.0*M_PI*radius);
		const float ratio = votes/(float)perimeter;

		int x_min = c_x - 5; if( x_min <  0       ) x_min = 0;
		int x_max = c_x + 5; if( x_max >= width   ) x_max = width-1;
		int y_min = c_y - 5; if( y_min <  0       ) y_min = 0;
		int y_max = c_y + 5; if( y_max >= height  ) y_max = height-1;
		int r_min = c_r - 5; if( r_min <  0       ) r_min = 0;
		int r_max = c_r + 5; if( r_max >= r_range ) r_max = r_range-1;

		for( int y = y_min; y <= y_max; y++ )
		{
			for( int x = x_min; x <= x_max; x++ )
			{
				const int i_tmp = (spaces_offset + x) + spaces_width*(y + spaces_offset);

				for( int r = r_min; r <= r_max; r++ )
					spaces[ r ][ i_tmp ] = 0;
			}
		}

		if( ratio >= tolerance )
		{
			printf( "cercle en [%d,%d] de rayon %d (p=%d), ratio=%.2f,nombre votants: %d\n", c_x, c_y, radius, perimeter, ratio, votes );

			DrawCircle( dest, c_x, c_y, radius );

			number--;
		}
		else
		{
			printf( "IGNORE cercle en [%d,%d] de rayon %d (p=%d), ratio=%.2f,nombre votants: %d\n", c_x, c_y, radius, perimeter, ratio, votes );


		}
	}

	for( int i = 0; i < r_range; i++ )
	{
		delete [] spaces[i];
	}
	delete [] spaces;
}




// Fonctions locales
void VotePoint( int x, int y, int* data, int width, int height )
{
	data[ x + width*y ]++;
}
void VoteCircle( int x0, int y0, int radius, int* data, int width, int height )
{
	int f = 1 - radius;
	int ddF_x = 0;
	int ddF_y = -2 * radius;
	int x = 0;
	int y = radius;
 
	VotePoint( x0, y0 + radius, data, width, height );
	VotePoint( x0, y0 - radius, data, width, height );
	VotePoint( x0 + radius, y0, data, width, height );
	VotePoint( x0 - radius, y0, data, width, height );
 
	while(x < y) 
	{
		if(f >= 0) 
		{
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x + 1;
		VotePoint( x0 + x, y0 + y, data, width, height );
		VotePoint( x0 - x, y0 + y, data, width, height );
		VotePoint( x0 + x, y0 - y, data, width, height );
		VotePoint( x0 - x, y0 - y, data, width, height );
		VotePoint( x0 + y, y0 + x, data, width, height );
		VotePoint( x0 - y, y0 + x, data, width, height );
		VotePoint( x0 + y, y0 - x, data, width, height );
		VotePoint( x0 - y, y0 - x, data, width, height );
	}
}

void DrawPoint( Image* im, int x, int y )
{
	if( x > 0 && x < (int)im->GetWidth() && y > 0 && y < (int)im->GetHeight() )
	{
		im->GetPixel( x, y )[ 0 ] = 255;
	}
}
void DrawCircle( Image* im, int x0, int y0, int radius )
{
	int f = 1 - radius;
	int ddF_x = 0;
	int ddF_y = -2 * radius;
	int x = 0;
	int y = radius;
 
	DrawPoint( im, x0, y0 + radius );
	DrawPoint( im, x0, y0 - radius );
	DrawPoint( im, x0 + radius, y0 );
	DrawPoint( im, x0 - radius, y0 );
 
	while(x < y) 
	{
		if(f >= 0) 
		{
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x + 1;
		DrawPoint( im, x0 + x, y0 + y );
		DrawPoint( im, x0 - x, y0 + y );
		DrawPoint( im, x0 + x, y0 - y );
		DrawPoint( im, x0 - x, y0 - y );
		DrawPoint( im, x0 + y, y0 + x );
		DrawPoint( im, x0 - y, y0 + x );
		DrawPoint( im, x0 + y, y0 - x );
		DrawPoint( im, x0 - y, y0 - x );
	}
}
