#include "hough.h"
#include "hough_utils.h"


#include <cstdio>
#include <cstring>




void ComputeLineHough( int x, int y, int* space, int space_width, int theta_max, int rho_max, int sign )
{
	for( int theta = 0; theta < theta_max; theta++ )
	{
		const double theta_rad = (double)theta * ( M_PI/180.0 );

		const int rho = (int)( x * cos(theta_rad) + y * sin(theta_rad) );

		if( rho >= 0 && rho <= rho_max )
		{
			const uint index = theta + space_width*(rho_max-rho);

			space[ index ] += sign;
		}
	}
}


void Hough::GetLines( Image* dest, int number )
{
	const int width  = m_source.GetWidth();
	const int height = m_source.GetHeight();

	PrepareDestination( dest, &m_source );

	// Initialisations
	const int theta_max = 180;
	const int rho_max   = (int)sqrt( (double)width*width + height*height );

	const int space_width        = theta_max + 1;
	const int space_height       = rho_max   + 1;
	const int space_pixel_number = space_width * space_height;

	int* space = new int[ space_pixel_number ];
	memset( space, 0, sizeof(*space) * space_pixel_number );

	// Calcul des votants
	for( int y = 0; y < height; y++ )
	{
		for( int x = 0; x < width; x++ )
		{
			if( m_source.GetPixel( x, y )[ 0 ] > 127 )
			{
				ComputeLineHough( x, y, space, space_width, theta_max, rho_max, 1 );
			}
		}
	}

	/*{
		Image image;
		image.Create( space_width, space_height, 1 );
		SaveToImage( &image, space, space_width, space_height );
		image.SavePNG( "hough_lignes.png" );
	}*/

	// Extraction des droites
	while( number-- )
	{
		// On recupere le max de l'espace de Hough
		int index_max = 0;
		for( int i = 1; i < space_pixel_number; i++ )
		{
			if( space[ index_max ] < space[ i ] )
			{
				index_max = i;
			}
		}

		const int votes = space[ index_max ];

		if( votes < 2 )
			break;

		const int theta = index_max % space_width;
		const int rho   = space_height - (index_max / space_width);

		space[ index_max ] = 0;

		// On trace la droite
		if( theta == 0 )
		{
			// Verticale
			const double t = theta * ( M_PI/180.0 );
			const int    x = (int)( rho/cos(t) ) - 1;

			printf( "droite verticale en x = %d (%d votants)\n", x, votes );

			for( int y = 0; y < height; y++ )
			{
				dest->GetPixel( x, y )[ 0 ] = 255;

				// Suppression de la droite dans l'espace
				ComputeLineHough( x, y, space, space_width, theta_max, rho_max, -1 );
			}
		}
		else if( theta == 90 )
		{
			// Horizontale
			const double t = theta * ( M_PI/180.0 );
			const int    y = (int)( rho / sin(t) );

			printf( "droite horizontale en y = %d (%d votants)\n", y, votes );

			for( int x = 0; x < width; x++ )
			{
				dest->GetPixel( x, y )[ 0 ] = 255;

				// Suppression de la droite dans l'espace
				ComputeLineHough( x, y, space, space_width, theta_max, rho_max, -1 );
			}
		}
		else
		{
			int x_good, y_good;

			const double t = (double)theta * ( M_PI/180.0 );
			const double r = (double)rho;

			for( int x = 0; x < width; x++ )
			{
				const int y = (int)( ( r - x*cos(t) ) / sin(t) );

				if( y >= 0 && y < height )
				{
					dest->GetPixel( x, y )[ 0 ] = 255;

					x_good = x;
					y_good = y;

					// Suppression de la droite dans l'espace
					ComputeLineHough( x, y, space, space_width, theta_max, rho_max, -1 );
				}
			}

			for( int y = 0; y < height; y++ )
			{
				const int x = (int)( r/cos(t) - y*tan(t) );

				if( x >= 0 && x < width )
				{
					dest->GetPixel( x, y )[ 0 ] = 255;

					// Suppression de la droite dans l'espace
					ComputeLineHough( x, y, space, space_width, theta_max, rho_max, -1 );
				}
			}

			printf( "droite passant par [%d,%d] et d'angle %d (%d votants)\n", x_good, y_good, theta, votes );
		}
	}

	delete [] space;
}
