#include "kmeans.h"

#include <algorithm>
#include <vector>
#include <float.h>
#include <assert.h>
#include <math.h>

class Point
{
public:
	Point()									{ this->x = 0; this->y = 0; }
	Point( unsigned int x, unsigned int y )	{ this->x = x; this->y = y; }
	unsigned int x, y;
};

class Group
{
public:
	Group() {}
	Group( Point c ) { center = c; }
	Point center;
	std::vector< Point > pixels;
};

int random( int min, int max )
{
	return rand()%(max-min) + min;
}

float distanceColor( const unsigned char* c1, const unsigned char* c2, const unsigned int chan_number )
{
	int square_sum = 0;
	for( unsigned int k = 0 ; k < chan_number ; k++ )
	{
		square_sum += (c1[k]-c2[k])*(c1[k]-c2[k]);
	}
	return sqrtf( (float)square_sum );
}

float distancePixelsLine( const Image& image, int x0, int y0, int x1, int y1 )
{
	float distance = 0.0f;

	bool steep = std::abs(y1 - y0) > std::abs(x1 - x0) ;
	if( steep )
	{
		std::swap( x0, y0 );
		std::swap( x1, y1 );
	}
	if( x0 > x1 )
	{
		std::swap( x0, x1 );
		std::swap( y0, y1 );
	}
	
	int deltax = x1 - x0 ;
	int deltay = abs(y1 - y0) ;
	int error = -(deltax + 1) / 2 ;
	int ystep ;
	int y = y0 ;
	ystep = ( y0 < y1 ? 1 : -1 );
	for( int x = x0 ; x < x1 ; x++ )
	{
		if( steep ) distance += distanceColor( image.GetPixel(y,x), image.GetPixel(x1,y1), image.GetChanNumber() );
		else distance += distanceColor( image.GetPixel(x,y), image.GetPixel(x1,y1), image.GetChanNumber() );
		error += deltay ;
		if( error >= 0 )
		{
			y += ystep ;
			error -= deltax ;
		}
	}

	return distance;
}
float distancePixelsColor( const Image& image, int x0, int y0, int x1, int y1, float delta = 0.5f )
{
	float dist_color = distanceColor( image.GetPixel(x0,y0), image.GetPixel(x1,y1), image.GetChanNumber() );
	float dist = sqrtf( (float)((x0-x1)*(x0-x1)+(y0-y1)*(y0-y1)) );
	return dist_color*delta + dist*(1.0f-delta) ;
}

Image kMeans( const Image& image, unsigned int groups_number, unsigned int max_iterations, float delta, DistType dist_type )
{
	Group* groups = new Group[groups_number];
	//Tirage alatoire d'un centre par groupe
	for( unsigned int group_index = 0 ; group_index < groups_number ; group_index++ )
	{
		groups[group_index].center.x = random( 0, image.GetWidth() );
		groups[group_index].center.y = random( 0, image.GetHeight() );
		//TODO meilleur tirage alatoire
	}

	while( max_iterations-- )
	{
		for( unsigned int group_index = 0 ; group_index < groups_number ; group_index++ )
		{
			groups[group_index].pixels.clear();
		}

		for( unsigned int i = 0 ; i < image.GetHeight() ; i++ )
		{
			for( unsigned int j = 0 ; j < image.GetWidth() ; j++ )
			{
				const unsigned char* ptr_im = image.GetPixel( j, i );

				// Attribution du pixel au centre le plus proche
				unsigned int group_min_index = groups_number;
				float distance_min = FLT_MAX;
				for( unsigned int group_index = 0 ; group_index < groups_number ; group_index++ )
				{
					float distance;
					switch( dist_type )
					{
					case COLOR_TYPE :
						distance = distancePixelsColor( image, j, i, groups[group_index].center.x, groups[group_index].center.y, delta );
						break;
					case LINE_TYPE :
						distance = distancePixelsLine( image, j, i, groups[group_index].center.x, groups[group_index].center.y );
						break;
					}
					if( distance < distance_min )
					{
						distance_min = distance;
						group_min_index = group_index;
					}
				}
				assert( group_min_index != groups_number );
				groups[ group_min_index ].pixels.push_back( Point( j, i ) );
			}
		}

		//Calcul des nouveaux centres
		for( unsigned int group_index = 0 ; group_index < groups_number ; group_index++ )
		{
			const unsigned int size = groups[group_index].pixels.size();
			if( size > 0 )
			{
				unsigned int sum_x = 0, sum_y = 0;
				for( unsigned int vector_index = 0 ; vector_index < size ; vector_index++ )
				{
					sum_x += groups[group_index].pixels[vector_index].x;
					sum_y += groups[group_index].pixels[vector_index].y;
				}
				groups[group_index].center.x = sum_x / size;
				groups[group_index].center.y = sum_y / size;
			}
		}
	}

	//Ecriture rsultat dans image_result
	Image image_result( image.GetWidth(), image.GetHeight(), image.GetChanNumber() );

	unsigned int* sum_color = new unsigned int[ image.GetChanNumber() ];
	for( unsigned int group_index = 0 ; group_index < groups_number ; group_index++ )
	{
		const unsigned int size = groups[group_index].pixels.size();
		if( size > 0 )
		{
			for( unsigned int k = 0 ; k < image.GetChanNumber() ; k++ ) sum_color[k] = 0;
			for( unsigned int vector_index = 0 ; vector_index < size ; vector_index++ )
			{
				const unsigned char* ptr_im = image.GetPixel( groups[group_index].pixels[vector_index].x, groups[group_index].pixels[vector_index].y );
				for( unsigned int k = 0 ; k < image.GetChanNumber() ; k++ ) sum_color[k] += ptr_im[k];
			}
			for( unsigned int k = 0 ; k < image.GetChanNumber() ; k++ ) sum_color[k] /= size;
			for( unsigned int vector_index = 0 ; vector_index < size ; vector_index++ )
			{
				unsigned char* ptr_im = image_result.GetPixel( groups[group_index].pixels[vector_index].x, groups[group_index].pixels[vector_index].y );
				for( unsigned int k = 0 ; k < image.GetChanNumber() ; k++ ) ptr_im[k] = (unsigned char)sum_color[k];
			}
		}
	}
	delete[] sum_color;
	delete[] groups;

	return image_result;
}